/*
   Copyright (C) 2006, 2007 Clark Williams <williams@redhat.com>
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
   USA */
/*

  Isolate is a program used to move threads and interrupts off of selected
  processors. 

 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <ctype.h>
#include <unistd.h>
#include <getopt.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <sched.h>
#include <fcntl.h>

// the cpus to clear of processes and irqs
cpu_set_t cpus_to_clear;

// the available processors on the system
cpu_set_t available_cpus;

int verbose = 0;
int dryrun = 0;
int debugging = 0;
int reset = 0;
char *version = VERSION " (" __TIME__ ", " __DATE__ ")";

char *myname = NULL;

#define TRUE 1
#define FALSE 0

// command line options
struct option options [] = {
	{ "dryrun", no_argument, NULL, 'D' },
	{ "verbose", no_argument, NULL, 'v' },
	{ "debug", no_argument, NULL, 'd'},
	{ "version", no_argument, NULL, 'V'},
	{ "help", no_argument, NULL, 'h'},
	{ "reset", no_argument, NULL, 'r' },
	{ NULL, 0, NULL, 0},
};

void get_irqs(int **irqlist, int *nirqs);

// print usage message and exit
void
usage(void)
{
	fputs("usage: isolate [options] cpu1 [cpu2...cpuN]\n", stderr);
	fputs("    options:\n", stderr);
	fputs("		--help		print this message and exit\n", stderr);
	fputs("		--verbose	turn on verbose output\n", stderr);
	fputs("		--debug		turn on debugging output\n", stderr);
	fputs("		--dryrun	print the actions that would occur\n", stderr);
	fputs("		--reset         reset all threads to all cpus and exits\n", stderr);
	exit(0);
}

// handle command line options and return the index
// of the first non-option argument
int
process_command_line(int argc, char **argv)
{
	int opt;
	while ((opt = getopt_long_only(argc, argv, "+", options, NULL)) != -1) {
		switch (opt) {
		case '?':
		case 'h':
			usage();
		case 'v':
			verbose = 1;
			break;
		case 'd':
			debugging = 1;
			break;
		case 'V':
			puts(version);
			exit(0);
		case 'D':
			dryrun = 1;
			verbose = 1;
			break;
		case 'r':
			reset = 1;
			break;
		default:
			fprintf(stderr, "Unknown option %c\n", opt);
			usage();
		}
	}
	return optind;
}

// print out an informational message
void
info(char *fmt, ...)
{
	va_list ap;
	if (verbose) {
		puts("INFO: ");
		va_start(ap, fmt);
		vprintf(fmt, ap);
		va_end(ap);
	}
}

// print out an error and then exit
void
error(char *fmt, ...)
{
	va_list ap;
	fputs("ERROR: ", stderr);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fputs(strerror(errno), stderr);
	exit(errno);
}

// print out an error and then exit
void
warn(char *fmt, ...)
{
	va_list ap;
	fputs("WARNING: ", stderr);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}


// print a debug message
void
debug(char *fmt, ...)
{
	va_list ap;

	if (debugging) {
		fputs("DEBUG: ", stderr);
		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
	}
}


//
// find out how many cpus are available
// create a cpu_set_t of all available cpus 
// and return the total number of cpus
//

int
get_cpu_info(cpu_set_t *avail_cpus)
{
	DIR *dp = opendir("/sys/devices/system/cpu");
	struct dirent *de;
	int ncpus = 0;
	int cpu;

	if (dp == NULL)
		error("can't open /sys/devices/system/cpu");

	CPU_ZERO(avail_cpus);

	// look at all the entries and count the processor entries
	while ((de = readdir(dp))) {
		if ((strncmp(de->d_name, "cpu", 3) == 0) && 
			isdigit(de->d_name[3])) {
			ncpus++;
			cpu = strtol(&de->d_name[3], NULL, 10);
			debug("adding cpu %d to available cpu set\n", cpu);
			CPU_SET(cpu, avail_cpus);
		}
		
	}
	closedir(dp);
	return ncpus;
}

int
move_thread(int pid, cpu_set_t *cpus_to_clear)
{
	int i;
	cpu_set_t mask;
	int nmasked = 0;

	// get pid's current affinity
	if (sched_getaffinity(pid, sizeof(cpu_set_t), &mask)) {
		if (errno != ESRCH)
			warn("sched_getaffinity failed for pid: %d\n");
		return FALSE;
	}

	// remove any cpus specified by the cpus_to_clear set
	for (i = 0; i < CPU_SETSIZE; i++) {
		if (CPU_ISSET(i, cpus_to_clear) && CPU_ISSET(i, &mask)) {
			CPU_CLR(i, &mask);
			nmasked++;
		}
	}
	// if we need to change the affinity, do so
	if (nmasked) {
		debug("masking %d processors from pid %d\n", nmasked, pid);
		if (!dryrun &&
			sched_setaffinity(pid, sizeof(cpu_set_t), &mask)) {
			if (errno != ESRCH)
				warn("sched_setaffinity failed for pid: %d\n");
			return FALSE;
		}
		return TRUE;
	}
	return FALSE;
}

// check to see if pid is a kernel thread
// this is done by reading the value of /proc/<pid>/maps
// if the maps field is empty, it's a kernel thread
//
int
iskthread(pid_t pid)
{
	int fd;
	int retval;
	char path[50];
	char buffer[80];

	snprintf(path, sizeof(path), "/proc/%d/maps", pid);

	if ((fd = open(path, O_RDONLY)) < 0) {
		if (errno == ENOENT)
			return FALSE;
		error("open of %s failed\n", path);
	}
	if ((retval = read(fd, buffer, sizeof(buffer))) < 0)
		error("read of %s failed\n", path);
	close(fd);
	if (retval)
		return FALSE;
	return TRUE;
}

int
isnumber(char *name)
{
	char *ptr = name;
	int retval = 1;

	while(*ptr) {
		if (!isdigit(*ptr)) {
			retval = 1;
			break;
		}
		ptr++;
	}
	return retval;
}

int
set_irq_affinity(int irq, cpu_set_t *mask)
{
	int i;
	char irqpath[80];
	char affinity_buf[160];
	unsigned int *vector;
	int vector_len = 0;
	int len;
	char *ptr;
	FILE *fp;

	snprintf(irqpath, sizeof(irqpath), "/proc/irq/%d/smp_affinity", irq);
	
	fp = fopen(irqpath, "r");
	if (fp == NULL)
		return -1;
	if ((fgets(affinity_buf, sizeof(affinity_buf), fp)) == NULL) {
		fclose(fp);
		return -1;
	}
	fclose(fp);
	len = strlen(affinity_buf);
	for (ptr = affinity_buf; ptr < &affinity_buf[len]; ptr++)
		if (*ptr == ',') vector_len++;
	vector_len++;
	
	vector = alloca(sizeof(unsigned int) * vector_len);
	memset(vector, 0, sizeof(unsigned int) * vector_len);

	for (i = 0; i < CPU_SETSIZE; i++) {
		if (CPU_ISSET(i, mask)) {
			int bit = i % 32;
			int idx = vector_len - (i / 32 ) - 1;
			debug("adding cpu %d to cpuset for irq %d (idx:%d, bit:%d)\n", 
			      i, irq, idx, bit);
			vector[idx] |= (1 << bit);
		}
	}
	affinity_buf[0] = '\0';
	for (i = 0, ptr = affinity_buf; i < vector_len; i++) {
		len = sprintf(ptr, "%8.8x,", vector[i]);
		ptr += len;
	}
	// nuke the trailing comma
	len = strlen(affinity_buf);
	affinity_buf[len-1] = '\0';

	debug("new affinity mask for %d is %s\n", irq, affinity_buf);

	// write it back
	if ((fp = fopen(irqpath, "w")) == NULL)
		return -1;
	fputs(affinity_buf, fp);
	fclose(fp);
	return 0;

}      	

void
reset_affinity_masks(cpu_set_t *all_cpus)
{
	DIR *dp;
	struct dirent *de;
	pid_t pid;
	int irq, i;
	int *irqlist, nirqs;

	// reset threads
	if ((dp = opendir("/proc")) == NULL)
		error("Failed to open /proc\n");

	while ((de = readdir(dp))) {
		if (de->d_type != DT_DIR)
			continue;
		if (!isnumber(de->d_name))
			continue;
		pid = strtol(de->d_name, NULL, 10);
		if (iskthread(pid)) {
			debug("skipping kthread %d in reset\n", pid);
			continue;
		}
		sched_setaffinity(pid, sizeof(cpu_set_t), all_cpus);
	}
	closedir(dp);

	// reset irqs
	get_irqs(&irqlist, &nirqs);
	for (i = 0; i < nirqs; i++) {
		irq = irqlist[i];
		set_irq_affinity(irq, all_cpus);
	}
	free(irqlist);
}

// scan all pids trying to move them off the 
// cpus_to_clear set
void
process_threads(cpu_set_t *cpus_to_clear)
{
	DIR *dp;
	struct dirent *de;
	pid_t pid;
	int moved = 0;
	int changed = 0;

retry:
	dp = opendir("/proc");
	if (dp == NULL)
		error("failed to open /proc\n");

	// read through /proc pid entries and
	// move any user-space threads off the
	// specified processors
	while ((de = readdir(dp))) {

		// skip non-directory entries
		if (de->d_type != DT_DIR)
			continue;

		// skip non-pid-like entries
		if (!isnumber(de->d_name))
			continue;

		// convert name to a pid value
		pid = strtol(de->d_name, NULL,  10);

		// verify it's not a kernel thread
		if (iskthread(pid))
			continue;

		// move it
		if (move_thread(pid, cpus_to_clear)) {
			moved++;
			changed++;
		}
	}
	closedir(dp);

	if (changed) {
		info("moved %d threads, making another pass\n", changed);
		changed = 0;
		goto retry;
	}

	info("moved %d tasks\n", moved);
}


void
get_irqs(int **irqlist, int *nirqs)
{
	FILE *fp = fopen("/proc/interrupts", "r");
	char buffer[80];
	char *ptr;
	int irqbufsz = 100;
	int *irqs = alloca(irqbufsz*sizeof(int));
	int num_irqs = 0;

	if (fp == NULL)
		error("failed to open /proc/interrupts\n");

	// get the first line which is just the cpu header line
	if (fgets(buffer, sizeof(buffer), fp) == NULL)
		error("unexpected EOF reading /proc/interrupts\n");

	while (fgets(buffer, sizeof(buffer), fp)) {
		ptr = strtok(buffer, " :");
		if (ptr == NULL)
			error("strtok failed on %s\n", buffer);
		// check for insane amount of irqs
		if (num_irqs >= irqbufsz)
			error("more than %d irqs encountered!\n", irqbufsz);

		// if the first token isn't numeric, then bail out
		if (!isdigit(ptr[0]))
			break;
		irqs[num_irqs++] = strtol(ptr, NULL, 10);
	}
	fclose(fp);

	*irqlist = malloc(sizeof(int) * num_irqs);
	if (*irqlist == NULL)
		error("malloc failed for %d bytes\n", sizeof(int) * num_irqs);
	memcpy(*irqlist, irqs, sizeof(int) * num_irqs);
	*nirqs = num_irqs;
}


// move the specified IRQ off the input processor(s)
int
move_irq(int irq, cpu_set_t *cpus_to_clear)
{
	int i;
	int fd;
	int retval;
	int vector_len = 0;
	int idx;
	int maxcpu;
	int nmasked = 0;
	char irqpath[80];
	char affinity_buf[80];
	char *ptr;
	unsigned int *affinity_vector;

	debug("processing irq %d\n", irq);

	snprintf(irqpath, sizeof(irqpath), "/proc/irq/%d/smp_affinity", irq);
	if ((fd = open(irqpath, O_RDONLY)) < 0) {
		warn("failed to open %s\n", irqpath);
		return FALSE;
	}

	if ((retval = read(fd, affinity_buf, sizeof(affinity_buf))) < 0) {
		warn("error reading %s\n", irqpath);
		return FALSE;
	}

	close(fd);

	debug("affinity buffer for irq%d: %s\n", irq, affinity_buf);
	for (i = 0; i < strlen(affinity_buf); i++)
		if (affinity_buf[i] == ',')
			vector_len++;
	vector_len++;
	affinity_vector = alloca(vector_len * sizeof(unsigned int));
	
	ptr = strtok(affinity_buf, ",");
	for (i = 0; ptr && i < vector_len; i++) {
		affinity_vector[i] = strtol(ptr, NULL, 16);
		ptr = strtok(NULL, ",");
	}
	debug("affinity value had %d 32-bit elements\n", vector_len);
	if (debugging) {
		for (i=0; i < vector_len; i++)
			debug("affinity_vector[%d]: %08.8x\n", i, affinity_vector[i]);
	}

	
	// step through all the processors, masking off anything set in the
	// cpus_to_clear set.
	// this is a little convoluted because the vector of interrupts
	// is reversed (with lsb being in the largest index of the vector)
	maxcpu = vector_len * 32 < CPU_SETSIZE ? vector_len * 32 : CPU_SETSIZE;
	debug("locating cpu bit in irq vector\n");
	for (i = 0; i < maxcpu; i++) {
		if (CPU_ISSET(i, cpus_to_clear)) {
			int bit = i % 32;
			idx = vector_len - (i / 32) - 1;
			debug("cpu %d falls in vector idx %d as bit %d\n", i, idx, bit);
			affinity_vector[idx] &= ~(1 << bit);
			nmasked++;
		}
	}
	if (nmasked) {
		// now we need to reassemble the character vector
		char buf[9];
		int len;

		affinity_buf[0] = '\0';
		for (i = 0; i < vector_len; i++) {
			snprintf(buf, sizeof(buf), "%8.8x", affinity_vector[i]);
			strcat(affinity_buf, buf);
			len = strlen(affinity_buf);
			affinity_buf[len] = ',';
			affinity_buf[len+1] = '\0';
		}
		len = strlen(affinity_buf);

		// nuke the traling comma
		affinity_buf[--len] = '\0';
		debug("new affinity buffer: %s\n", affinity_buf);

		if (!dryrun) {
			if ((fd = open(irqpath, O_WRONLY)) < 0) {
				warn("failed to open %s for writing\n", irqpath);
				return FALSE;
			}
			// don't fail a write
			retval = write(fd, affinity_buf, len);
			close(fd);
			info("set new affinity for irq %d\n", irq);
		}
		return TRUE;
	}
	return FALSE;
}

void
process_irqs(cpu_set_t *cpus_to_clear)
{
	int i;
	int *irqlist, nirqs;
	int num_moved = 0;

	get_irqs(&irqlist, &nirqs);

	if (nirqs == 0)
		error("failed to get irq information");

	for (i=0; i <  nirqs; i++)
		if (move_irq(irqlist[i], cpus_to_clear))
			num_moved++;
	free(irqlist);
	info("moved %d irqs\n", num_moved);
}

void
check_for_irqbalance(void)
{
	if (access("/var/lock/subsys/irqbalance", F_OK) == 0) {
		warn("\nThe IRQ Balance daemon is listed as running!\n"
		     "It may change anything you do with isolate!\n"
		     "Please stop irqbalance before attempting to tune interrupts\n\n");
	}
}


int
main(int argc, char *argv[])
{
	int i, cpu;
	int first_arg = process_command_line(argc, argv);
	int num_cpus = get_cpu_info(&available_cpus);
	int num_to_clear = 0;

	myname = basename(argv[0]);

	if (getuid() != 0)
		error("%s must run as root!\n", myname);

	check_for_irqbalance();

	if (reset) {
		reset_affinity_masks(&available_cpus);
		exit(0);
	}


	info("%d processor%s available\n", num_cpus, num_cpus > 1 ? "s" : "");

	// gather up the processor numbers on the command line
	// and put them into a cpu set
	CPU_ZERO(&cpus_to_clear);
	for (i = first_arg; i < argc; i++) {
		cpu = strtol(argv[i], NULL, 10);
		if (cpu > num_cpus || !CPU_ISSET(cpu, &available_cpus))
			error("Invalid cpu number %d (max: %d)\n", cpu, num_cpus);
		CPU_SET(cpu, &cpus_to_clear);
		num_to_clear++;
	}

	if (num_to_clear == 0)
		usage();

	info("%d processors to clear\n", num_to_clear);

	if (num_to_clear >= num_cpus)
		error("can't clear all cpus!\n");

	// boot threads off the selected processors
	process_threads(&cpus_to_clear);

	// now move irqs off
	process_irqs(&cpus_to_clear);
	
	exit(0);
}
